"""
Code Generation
----------------
"""
import logging
import pathlib
import pprint
import typing

from graphql import (
    DocumentNode,
    Node,
)

from cannula.schema import concat_documents
from cannula.types import Argument, Directive, Field, FieldType, ObjectType

LOG = logging.getLogger(__name__)

TYPES = {
    "Boolean": "bool",
    "Float": "float",
    "ID": "str",
    "Int": "int",
    "String": "str",
}
VALUE_FUNCS = {
    "boolean_value": lambda value: value in ["true", "True"],
    "int_value": lambda value: int(value),
    "float_value": lambda value: float(value),
}


def parse_description(obj: dict) -> typing.Optional[str]:
    raw_description = obj.get("description") or {}
    return raw_description.get("value")


def parse_name(obj: dict) -> str:
    raw_name = obj.get("name") or {}
    return raw_name.get("value", "unknown")


def parse_value(obj: dict) -> typing.Any:
    kind = obj["kind"]
    value = obj["value"]

    if func := VALUE_FUNCS.get(kind):
        return func(value)

    return value


def parse_args(obj: dict) -> typing.List[Argument]:
    args: typing.List[Argument] = []
    raw_args = obj.get("arguments") or []
    for arg in raw_args:
        name = parse_name(arg)
        arg_type = FieldType(value=None, required=False)
        raw_type = arg.get("type")
        if raw_type:
            arg_type = parse_type(raw_type)
        value = None
        raw_value = arg.get("value")
        if raw_value:
            value = parse_value(raw_value)
        default = parse_default(arg)
        args.append(
            Argument(
                name=name,
                type=arg_type.value,
                required=arg_type.required,
                value=value,
                default=default,
            )
        )

    return args


def parse_type(type_obj: dict) -> FieldType:
    required = False
    value = None

    if type_obj["kind"] == "non_null_type":
        required = True
        value = parse_type(type_obj["type"]).value

    if type_obj["kind"] == "list_type":
        list_type = parse_type(type_obj["type"]).value
        value = f"typing.List[{list_type}]"

    if type_obj["kind"] == "named_type":
        name = type_obj["name"]
        value = name["value"]
        if value in TYPES:
            value = TYPES[value]
        else:
            value = f"{value}Type"

    return FieldType(value=value, required=required)


def parse_directives(field: typing.Dict[str, typing.Any]) -> typing.List[Directive]:
    directives: typing.List[Directive] = []
    for directive in field.get("directives", []):
        name = parse_name(directive)
        args = parse_args(directive)
        directives.append(Directive(name=name, args=args))
    return directives


def has_computed_directive(field: Field) -> bool:
    for directive in field.directives:
        if directive.name == "computed":
            return True

    return False


def parse_default(obj: typing.Dict[str, typing.Any]) -> typing.Any:
    default_value = obj.get("default_value")
    LOG.debug(default_value)
    if not default_value:
        return None

    return parse_value(default_value)


def parse_field(field: typing.Dict[str, typing.Any], parent: str) -> Field:
    name = field["name"]["value"]
    field_type = parse_type(field["type"])
    default = parse_default(field)
    directives = parse_directives(field)
    args = parse_args(field)
    func_name = f"{name}{parent}"

    return Field(
        name=name,
        value=field_type.value,
        func_name=func_name,
        description=parse_description(field),
        directives=directives,
        args=args,
        default=default,
        required=field_type.required,
    )


def parse_node(node: Node):
    details = node.to_dict()
    LOG.debug("%s: %s", node.kind, pprint.pformat(details))
    name = details.get("name", {}).get("value", "Unknown")
    raw_fields = details.get("fields", [])
    raw_description = details.get("description") or {}
    description = raw_description.get("value")
    directives: typing.Dict[str, typing.List[Directive]] = {}

    fields: typing.List[Field] = []
    for field in raw_fields:
        parsed = parse_field(field, parent=name)
        fields.append(parsed)
        if parsed.directives:
            directives[parsed.name] = parsed.directives

    return ObjectType(
        name=name, fields=fields, directives=directives, description=description
    )


def parse_schema(
    type_defs: typing.Iterable[typing.Union[str, DocumentNode]]
) -> typing.Dict[str, ObjectType]:
    document = concat_documents(type_defs)
    types: typing.Dict[str, ObjectType] = {}

    for definition in document.definitions:
        node = parse_node(definition)
        if node.name in types:
            types[node.name].fields.extend(node.fields)
            types[node.name].directives.update(node.directives)
        else:
            types[node.name] = node

    return types


computed_field_template = """\
    @abc.abstractmethod
    async def {field.name}(
        self,
        info: cannula.ResolveInfo,
{rendered_args}) -> typing.Awaitable[{field.value}]:
        ...

"""

required_field_template = """\
    {field.name}: {field.value}
"""

optional_field_template = """\
    {field.name}: typing.Optional[{field.value}] = {field.default!r}
"""

object_dict_fields = """\
    {field.name}: {field_type}
"""

object_template = """\
@dataclasses.dataclass
class {obj.name}TypeBase(abc.ABC):
    __typename = "{obj.name}"

{rendered_base_fields}
{rendered_computed}
class {obj.name}TypeDict(typing.TypedDict):
{rendered_dict_fields}

{obj.name}Type = typing.Union[{obj.name}TypeBase, {obj.name}TypeDict]
"""

function_args_template = """\
        {arg.name}: {arg_type}{default},
"""

function_template = """\
class {field.func_name}(typing.Protocol):
    def __call__(
        self,
        info: cannula.ResolveInfo,
{rendered_args}) -> typing.Awaitable[{field.value}]:
        ...
"""

operation_field_template = """\
    {field.name}: NotRequired[{field.func_name}]
"""

operation_template = """\
class RootType(typing.TypedDict):
{rendered_fields}"""

base_template = """\
from __future__ import annotations

import abc
import dataclasses
import typing

from typing_extensions import NotRequired

import cannula


{rendered_items}"""


def render_computed_field(field: Field) -> str:
    rendered_args = "".join(render_function_args(arg) for arg in field.args)
    rendered_args = f"{rendered_args}    "
    return computed_field_template.format(field=field, rendered_args=rendered_args)


def render_field(field: Field) -> str:
    if field.required:
        return required_field_template.format(field=field)
    return optional_field_template.format(field=field)


def render_dict_field(field: Field) -> str:
    field_type = field.value if field.required else f"NotRequired[{field.value}]"
    return object_dict_fields.format(field=field, field_type=field_type)


def render_object(obj: ObjectType) -> str:
    non_computed: typing.List[Field] = []
    computed: typing.List[Field] = []
    for field in obj.fields:
        if field.args or has_computed_directive(field):
            computed.append(field)
        else:
            non_computed.append(field)

    rendered_computed = "".join([render_computed_field(f) for f in computed])
    rendered_base_fields = "".join([render_field(f) for f in non_computed])
    rendered_dict_fields = "".join([render_dict_field(f) for f in obj.fields])
    return object_template.format(
        obj=obj,
        rendered_computed=rendered_computed,
        rendered_base_fields=rendered_base_fields,
        rendered_dict_fields=rendered_dict_fields,
    )


def render_function_args(arg: Argument) -> str:
    if arg.required:
        default = f" = {arg.default!r}" if arg.default else ""
        return function_args_template.format(
            arg=arg, arg_type=arg.type, default=default
        )

    arg_type = f"typing.Optional[{arg.type}]"
    default = f" = {arg.default!r}"

    return function_args_template.format(arg=arg, arg_type=arg_type, default=default)


def render_function(field: Field) -> str:
    rendered_args = "".join(render_function_args(arg) for arg in field.args)
    rendered_args = f"{rendered_args}    "
    return function_template.format(field=field, rendered_args=rendered_args)


def render_operation_field(field: Field) -> str:
    return operation_field_template.format(field=field)


def render_operation(fields: typing.List[Field]) -> str:
    rendered_fields = "".join([render_operation_field(f) for f in fields])
    return operation_template.format(
        rendered_fields=rendered_fields,
    )


def render_file(
    type_defs: typing.Iterable[typing.Union[str, DocumentNode]],
    path: pathlib.Path,
    dry_run: bool = False,
) -> None:
    parsed = parse_schema(type_defs)

    object_types: typing.List[ObjectType] = []
    operation_fields: typing.List[Field] = []
    for obj in parsed.values():
        if obj.name in ["Query", "Mutation", "Subscription"]:
            for field in obj.fields:
                operation_fields.append(field)
        else:
            object_types.append(obj)

    object_types.sort(key=lambda o: o.name)
    objects = [render_object(obj) for obj in object_types]

    operation_fields.sort(key=lambda f: f.name)
    operations = render_operation(operation_fields)
    functions = [render_function(field) for field in operation_fields]

    rendered_objects = "\n\n".join(objects + functions)
    rendered_items = f"{rendered_objects}\n\n{operations}"
    content = base_template.format(
        rendered_items=rendered_items,
    )

    if dry_run:
        LOG.info(f"DRY_RUN would produce: \n{content}")
        return

    with open(path, "w") as output:
        output.writelines(content)



"""
Context is how you can share information between resolvers. By default when
a request comes in cannula will create a context instance with the request
as an attribute. Then this is added to the GraphQLResolve info and passed
as the second argument to resolvers.

Example Resolver::

    @api.query
    def get_something(parent: Any, info: ResolveInfo[Context]):
        original_request = info.context.request
        if not can_access_something(original_request):
            raise AccessDenied("you do not have permission!")
        return get_something()

Context API Docs
----------------
"""

import typing
from graphql import GraphQLResolveInfo

C = typing.TypeVar("C")
R = typing.TypeVar("R")


class ResolveInfo(typing.Generic[C], GraphQLResolveInfo):
    """
    This class is strictly to help with type checking. You can use
    this as the `info` arg in a revolver to assist with type hints
    and code completion::

        import typing
        import cannula

        class CustomContext(cannula.Context):
            widgets = widget_datasource()

        @api.query
        def get_widgets(
            parent: typing.Any,
            info: cannual.ResolveInfo[CustomContext],
        ) -> typing.List[Widget]:
            # type checker will be able to verify `get_widgets`
            # has the correct return type `list[Widget]`
            return info.context.widgets.get_widgets()
    """

    context: C


class Context(typing.Generic[R]):
    """Default Context Base

    Subclasses should implement a handle_request method to provide any
    extra functionality they need.
    """

    request: R

    def __init__(self, request: R):
        self.request = self.handle_request(request)

    @classmethod
    def init(cls, request: R):
        return cls(request)

    def handle_request(self, request: R) -> R:
        return request



import typing

import pydantic


class GraphQLPayload(pydantic.BaseModel):
    """Model representing a GraphQL request body."""

    query: str
    variables: typing.Optional[typing.Dict[str, typing.Any]] = None
    operation: typing.Optional[str] = None



import collections
import logging
import typing

from graphql import GraphQLError


def format_errors(
    errors: typing.Optional[typing.List[GraphQLError]] = None,
    logger: typing.Optional[logging.Logger] = None,
    level: int = logging.DEBUG,
) -> dict:
    """Return a dict object of the errors.

    If there is a path(s) in the error then return a dict with the path
    as a key so that it is easier on the client side code to display the
    error with the correct data.
    """
    if not errors:
        return {}

    if logger is None:
        logger = logging.getLogger(__name__)

    formatted_errors: typing.Dict[str, typing.List] = collections.defaultdict(list)

    for err in errors:
        log_error(err, logger, level)
        error_formatted = err.formatted
        error_message = error_formatted.get("message")
        if err.path is not None:
            for path in err.path:
                if error_message not in formatted_errors[str(path)]:
                    formatted_errors[str(path)].append(error_message)

        if error_message not in formatted_errors["errors"]:
            formatted_errors["errors"].append(error_message)

    return formatted_errors


def log_error(
    error: GraphQLError,
    logger: logging.Logger,
    level: int,
):
    logger.log(level, f"{error}")
    if tb := error.__traceback__:
        while tb and tb.tb_next:
            tb = tb.tb_next
        logger.log(level, f"Excecution Context: {tb.tb_frame.f_locals!r}")



"""
.. _httpdatasource:

HTTP Data Source
================

.. note::

    This requires the http extras to be installed::

        pip install cannula[http]

"""

import asyncio
import logging
import types
import typing

import httpx

LOG = logging.getLogger("cannula.datasource.http")


# solves the issue of `cannot reuse already awaited coroutine`
def cacheable(f):
    def wrapped(*args, **kwargs):
        r = f(*args, **kwargs)
        return asyncio.ensure_future(r)

    return wrapped


class Request(typing.NamedTuple):
    url: str
    method: str
    body: typing.Any = None
    headers: typing.Dict = {}


class HTTPDataSource:
    """
    HTTP Data Source

    This is modeled after the apollo http datasource. It uses httpx to preform
    async requests to any remote service you wish to query.

    Properties:

    * `base_url`: Optional base_url to apply to all requests
    * `timeout`: Default timeout in seconds for requests (5 seconds)
    * `resource_name`: Optional name to use for `__typename` in responses.
    """

    # The base url of this resource
    base_url: typing.Optional[str] = None
    # A mapping of requests using the cache_key_for_request. Multiple resolvers
    # could attempt to fetch the same resource, using this we can limit to at
    # most one request per cache key.
    memoized_requests: typing.Dict[str, typing.Awaitable]

    # Timeout for an individual request in seconds.
    timeout: int = 5

    # Resource name for the type that this datasource returns by default this
    # will use the class name of the datasource.
    resource_name: typing.Optional[str] = None

    def __init__(
        self,
        request: typing.Any,
        client: typing.Optional[httpx.AsyncClient] = None,
    ):
        self.client = client or httpx.AsyncClient()
        # close the client if this instance opened it
        self._should_close_client = client is None
        self.request = request
        self.memoized_requests = {}
        self.assert_has_resource_name()

    def __del__(self):
        if self._should_close_client:
            LOG.debug(f"Closing httpx session for {self.resource_name}")
            asyncio.ensure_future(self.client.aclose())

    def assert_has_resource_name(self) -> None:
        if self.resource_name is None:
            self.resource_name = self.__class__.__name__

    def will_send_request(self, request: Request) -> Request:
        """Hook for subclasses to modify the request before it is sent.

        For example setting Authorization headers::

            def will_send_request(self, request):
                request.headers = self.request.headers
                return request
        """
        return request

    def cache_key_for_request(self, request: Request) -> str:
        return request.url

    def get_request_url(self, path: str) -> str:
        if path.startswith(("https://", "http://")):
            return path

        if self.base_url is not None:
            if path.startswith("/"):
                path = path[1:]

            if self.base_url.endswith("/"):
                return f"{self.base_url}{path}"

            return f"{self.base_url}/{path}"

        return path

    def did_receive_error(self, error: Exception, request: Request):
        """Handle errors from the remote resource"""
        raise error

    def convert_to_object(self, json_obj):
        json_obj.update({"__typename": self.resource_name})
        return types.SimpleNamespace(**json_obj)

    async def did_receive_response(
        self, response: httpx.Response, request: Request
    ) -> typing.Any:
        """Hook to alter the response from the server.

        example::

            async def did_receive_response(
                self, response: httpx.Response, request: Request
            ) -> typing.Any:
                response.raise_for_status()
                return Widget(**response.json())
        """
        response.raise_for_status()
        return response.json(object_hook=self.convert_to_object)

    async def get(self, path: str) -> typing.Any:
        """Preform a GET request

        :param path: path of the request
        """
        return await self.fetch("GET", path)

    async def post(self, path: str, body: typing.Any) -> typing.Any:
        """Preform a POST request

        :param path: path of the request
        :param body: body of the request
        """
        return await self.fetch("POST", path, body)

    async def patch(self, path: str, body: typing.Any) -> typing.Any:
        """Preform a PATCH request

        :param path: path of the request
        :param body: body of the request
        """
        return await self.fetch("PATCH", path, body)

    async def put(self, path: str, body: typing.Any) -> typing.Any:
        """Preform a PUT request

        :param path: path of the request
        :param body: body of the request
        """
        return await self.fetch("PUT", path, body)

    async def delete(self, path: str) -> typing.Any:
        """Preform a DELETE request

        :param path: path of the request
        """
        return await self.fetch("DELETE", path)

    async def fetch(
        self, method: str, path: str, body: typing.Any = None
    ) -> typing.Any:
        url = self.get_request_url(path)

        request = Request(url, method, body)

        request = self.will_send_request(request)

        cache_key = self.cache_key_for_request(request)

        @cacheable
        async def process_request():
            try:
                response = await self.client.request(
                    request.method,
                    request.url,
                    json=request.body,
                    headers=request.headers,
                    timeout=self.timeout,
                )
            except Exception as exc:
                return self.did_receive_error(exc, request)
            else:
                return await self.did_receive_response(response, request)

        if request.method == "GET":
            promise = self.memoized_requests.get(cache_key)
            if promise is not None:
                LOG.debug(f"cache found for GET '{cache_key}'")
                return await promise

            self.memoized_requests[cache_key] = process_request()
            LOG.debug(f"cache set for GET '{cache_key}'")

            return await self.memoized_requests[cache_key]
        else:
            self.memoized_requests.pop(cache_key, None)
            return await process_request()



"""
Schema Utilities
----------------
"""
import logging
import pathlib
import typing
import itertools

from graphql import (
    GraphQLSchema,
    DocumentNode,
    TypeDefinitionNode,
    parse,
    build_ast_schema,
    extend_schema,
    concat_ast,
    is_type_extension_node,
    is_type_system_extension_node,
    is_type_definition_node,
)

LOG = logging.getLogger(__name__)
QUERY_TYPE = parse("type Query { _empty: String }")
MUTATION_TYPE = parse("type Mutation { _empty: String }")
COMPUTED_DIRECTIVE = parse("directive @computed(weight: Int = 1) on FIELD_DEFINITION")


def extract_extensions(ast: DocumentNode) -> DocumentNode:
    type_extensions = filter(is_type_extension_node, ast.definitions)
    system_extensions = filter(is_type_system_extension_node, ast.definitions)
    extensions = itertools.chain(type_extensions, system_extensions)
    return DocumentNode(definitions=list(extensions))


def assert_has_query_and_mutation(ast: DocumentNode) -> DocumentNode:
    """Assert that schema has query and mutation types defined.

    The schema is pretty much useless without them and rather than causing
    an error we'll just add in an empty one so they can be extended.
    """
    object_kinds: typing.List[TypeDefinitionNode] = filter(
        is_type_definition_node, ast.definitions
    )  # type: ignore
    object_definitions = [node.name.value for node in object_kinds]
    has_mutation_definition = "Mutation" in object_definitions
    has_query_definition = "Query" in object_definitions

    if not has_mutation_definition:
        LOG.debug("Adding default empty Mutation type")
        ast = concat_ast([ast, MUTATION_TYPE])

    if not has_query_definition:
        LOG.debug("Adding default empty Query type")
        ast = concat_ast([ast, QUERY_TYPE])

    LOG.debug("Adding computed directive")
    ast = concat_ast([ast, COMPUTED_DIRECTIVE])

    return ast


def maybe_parse(type_def: typing.Union[str, DocumentNode]):
    if isinstance(type_def, str):
        return parse(type_def)
    return type_def


def concat_documents(
    type_defs: typing.Iterable[typing.Union[str, DocumentNode]],
) -> DocumentNode:
    document_list = [maybe_parse(type_def) for type_def in type_defs]

    return concat_ast(document_list)


def build_and_extend_schema(
    type_defs: typing.Iterable[typing.Union[str, DocumentNode]],
) -> GraphQLSchema:
    """
    Build and Extend Schema

    When splitting schema into multiple files it is helpful to be able to
    extend an existing type. This is most commonly done with the `Query`
    and `Mutation` types.

    For example in one schema file you add a `Query` type::

        type Book {
            name: String
        }

        type Query {
            books: [Books]
        }

    Now in another schema file you can extend the `Query`::

        type Movie {
            name: String
        }

        extend type Query {
            movies: [Movie]
        }

    :param type_defs: list of schema or document nodes
    """
    ast_document = concat_documents(type_defs)

    ast_document = assert_has_query_and_mutation(ast_document)

    schema = build_ast_schema(ast_document)

    extension_ast = extract_extensions(ast_document)

    if extension_ast.definitions:
        LOG.debug("Extending schema")
        schema = extend_schema(
            schema, extension_ast, assume_valid=True, assume_valid_sdl=True
        )

    return schema


def load_schema(
    directory: typing.Union[str, pathlib.Path]
) -> typing.List[DocumentNode]:
    """
    Load Schema

    This utility will load schema from a directory or a single pathlib.Path

    :param directory: Directory to load schema files from
    """
    if isinstance(directory, str):
        LOG.debug(f"Converting str {directory} to path object")
        directory = pathlib.Path(directory)

    if directory.is_file():
        LOG.debug(f"loading schema from file: {directory}")
        with open(directory.absolute()) as graphfile:
            return [parse(graphfile.read())]

    def find_graphql_files():
        LOG.debug(f"Checking for graphql files to load in: '{directory}'")
        for graph in directory.glob("**/*.graphql"):
            LOG.debug(f"loading discovered file: {graph}")
            with open(graph) as graphfile:
                yield graphfile.read()

    return [parse(schema) for schema in find_graphql_files()]



"""
.. _mock-middleware:

MockMiddleware
==============

Wrapps an existing schema and resolvers to provide an easy way to automatically
mock the responses from Queries, Mutations or Subscriptions. You can choose
to mock all resolvers or you could selectively provide mock objects to replace
only a handful of calls while preserving the existing resolvers.

This is really useful for end to end testing, as you can simply add a header
to the request with the mock objects you wish to replace. This way the entire
stack is validated down to the resolver level.

Example Usage
-------------

You can add the middleware in a couple different ways. If you are using the
cannula API you can add this in the middleware list in the contructor::

    import cannula
    from cannula.middleware import MockMiddleware

    api = cannula.API(
        __name__,
        SCHEMA,
        middleware = [
            MockMiddleware(
                mock_all=True,
                mock_objects={
                    'Query': {
                        'field_to_fake': 'mock_value'
                    },
                    'String': 'I will be used for any String type',
                }
            ),
        ]
    )

Or you can use this with just the `graphql-core-next` library like::

    from cannula.middleware import MockMiddleware
    from graphql import graphql

    graphql(
        schema=SCHEMA,
        query=QUERY,
        middleware=[
            MockMiddleware(
                mock_all=True,
                mock_objects={
                    'String': 'just like before'
                }
            ),
        ],
    )

Example Using `X-Mock-Objects` Header
-------------------------------------

Most testing frameworks have a way to add headers to requests that you are
testing. Usually this done for authentication, but we are going to abuse this
functionality to tell the server what data to respond with. Here is an
example using Cypress.io:

.. code-block:: javascript

    var resourceMock = JSON.stringify({
        "Resource": {
            "name": "Mocky",
            "id": "12345"
        }
    });

    describe('Test Resource View', function() {
        it('Renders the resource correctly', function() {
            cy.server({
                onAnyRequest: function(route, proxy) {
                    proxy.xhr.setRequestHeader(
                        'X-Mock-Objects', resourceMock
                    );
                }
            });
            cy.visit('http://localhost:8000/resource/view/');
            cy.get('.resource').within(() => {
                cy.get('.name').should('equal', 'Mocky');
                cy.get('.id').should('equal', '12345');
            });
        })
    });

The key difference here from mocking the request is that we are actually
making the request to the server at localhost and we know that the routes are
setup correctly. We can freely change the payload and the urls that the actual
code uses and this test will continue to function. That is, unless we break
the contract of this page and those routes no longer respond with the data
we are testing for. Since we are not actually mocking the request or the
response breaking changes will be realized by this test.

Another reason why this pattern is great is that we are not testing against
a mock server that is specifically setup to respond to our request. While that
will work just fine the mocks are hidden from the tests in some other file.
Changing that file is complicated especially if it is used in multiple tests.

Mock Middleware API
-------------------
"""

import inspect
import json
import logging
import random
import typing
import uuid

from graphql import (
    GraphQLList,
    GraphQLType,
    get_named_type,
)

MockObjectTypes = typing.Union[typing.Callable, str, int, float, bool, dict, list]

LOG = logging.getLogger(__name__)

ADJECTIVES: typing.List[str] = [
    "imminent",
    "perfect",
    "organic",
    "elderly",
    "dapper",
    "reminiscent",
    "mysterious",
    "trashy",
    "workable",
    "flaky",
    "offbeat",
    "spooky",
    "thirsty",
    "stereotyped",
    "wild",
    "devilish",
    "quarrelsome",
    "dysfunctional",
]
NOUNS: typing.List[str] = [
    "note",
    "yak",
    "hammer",
    "cause",
    "price",
    "quill",
    "truck",
    "glass",
    "color",
    "ring",
    "trees",
    "window",
    "letter",
    "seed",
    "sponge",
    "pie",
    "mass",
    "table",
    "plantation",
    "battle",
]

LIST_LENGTH_KEY: str = "__list_length"
DEFAULT_MOCKS: typing.Dict[str, MockObjectTypes] = {
    "String": lambda: f"{random.choice(ADJECTIVES)}-{random.choice(NOUNS)}",
    "Int": lambda: random.randint(4, 999),
    "Float": lambda: random.randint(5, 999) * random.random(),
    "Boolean": lambda: random.choice([True, False]),
    "ID": lambda: str(uuid.uuid4()),
    # The default number of mock items to return when results are a list.
    LIST_LENGTH_KEY: lambda: random.randint(3, 6),
}


class SuperDict(dict):
    """Wraps a dictionary to provide attribute access.

    This is returned instead of a plain `dict` object to handle cases where
    the underlining resolvers where expecting an object returned. Since we
    allow setting up mocks via a JSON header this makes it possible to
    mock only the responses while preserving the rest of the resolvers attached
    to the schema.

        >>> mock = SuperDict({'foo': {'bar': {'baz': 42}}})
        >>> print(f"{mock.foo.bar.baz} == {mock['foo']['bar']['baz']}")
        42 == 42
    """

    def __init__(self, mapping: dict):
        super().__init__([], **mapping)
        for key, value in mapping.items():
            if isinstance(value, dict):
                value = SuperDict(value)
            setattr(self, key, value)


class MockObjectStore:
    """Stores and processes the mock objects to return correct results.

    The mock objects can be various types and instead of just returning
    the raw mock we first want to convert it to a better format. IE if the
    mock is callable we want to call it first and if the results or the
    mock are a dict, we need to wrap that in our SuperDict object.
    """

    __slots__ = ["mock_objects"]

    def __init__(self, mock_objects: typing.Dict[str, MockObjectTypes]):
        self.mock_objects = mock_objects

    def __contains__(self, key):
        return key in self.mock_objects

    def has_named_type(self, named_type_name: str) -> bool:
        return named_type_name in self.mock_objects

    def has_parent_field(self, parent_type_name: str, field_name: str) -> bool:
        return (
            parent_type_name in self.mock_objects
            and isinstance(self.mock_objects[parent_type_name], dict)
            and field_name
            in typing.cast(typing.Dict, self.mock_objects[parent_type_name])
        )

    def get(self, key: str, default: typing.Optional[typing.Any] = None) -> typing.Any:
        mock = self.mock_objects.get(key, default)
        if mock is None:
            return

        return self.return_results(mock)

    def get_parent_field(self, parent_type_name: str, field_name: str) -> typing.Any:
        parent = self.mock_objects.get(parent_type_name)

        mock = (
            parent.get(field_name)
            if isinstance(parent, dict)
            else getattr(parent, field_name, None)
        )

        if mock is None:
            return

        return self.return_results(mock)

    def return_results(self, mock: MockObjectTypes) -> typing.Any:
        results = mock() if callable(mock) else mock

        if isinstance(results, list):
            return [self.maybe_wrap(r) for r in results]

        return self.maybe_wrap(results)

    def maybe_wrap(self, results: typing.Any) -> typing.Any:
        if isinstance(results, dict):
            return SuperDict(results)
        return results


class MockMiddleware:
    """Mocking Middleware

    Initialize the MockMiddleware, you can choose to mock all types or just
    the ones provided by the `mock_objects` param. Alternatively you can
    provide the mock_objects via a request header.

    :param mock_objects: Mapping of return values for types.
    :param mock_all: Whether to mock all resolvers.
    :param mock_object_header: Customize the header used to set mock_objects.
    """

    def __init__(
        self,
        mock_objects: typing.Dict[str, MockObjectTypes] = {},
        mock_all: bool = True,
        mock_object_header: str = "X-Mock-Objects",
    ):
        self.mock_all = mock_all
        self.mock_object_header = mock_object_header
        self._mock_objects = DEFAULT_MOCKS.copy() if mock_all else {}
        self._mock_objects.update(mock_objects)

    def get_mocks(
        self, extra: typing.Dict[str, MockObjectTypes]
    ) -> typing.Optional[MockObjectStore]:
        mock_objects = self._mock_objects.copy()
        mock_objects.update(extra)

        if not mock_objects:
            return None

        return MockObjectStore(mock_objects)

    def get_mocks_from_headers(self, context: typing.Any) -> dict:
        try:
            mock_header = context.request.headers.get(self.mock_object_header)
            if not mock_header:
                return {}
        except Exception:
            return {}

        try:
            return json.loads(mock_header)
        except Exception:
            return {}

    async def run_next(self, _next, _resource, _info, **kwargs):
        results = _next(_resource, _info, **kwargs)

        if inspect.isawaitable(results):
            return await results
        return results

    async def resolve(self, _next, _resource, _info, **kwargs):
        mock_header = self.get_mocks_from_headers(_info.context)
        mock_objects = self.get_mocks(extra=mock_header)
        if not mock_objects:
            return await self.run_next(_next, _resource, _info, **kwargs)

        field_name = _info.field_name
        return_type = _info.return_type
        named_type = get_named_type(return_type)
        parent_type_name = _info.parent_type.name

        resolver_is_mocked = (
            self.mock_all
            or mock_objects.has_named_type(named_type.name)
            or mock_objects.has_parent_field(parent_type_name, field_name)
        )

        if not resolver_is_mocked:
            return await self.run_next(_next, _resource, _info, **kwargs)

        if mock_objects.has_parent_field(parent_type_name, field_name):
            return mock_objects.get_parent_field(parent_type_name, field_name)

        # Check if we previously resolved a mock object.
        field_value = (
            _resource.get(field_name)
            if isinstance(_resource, dict)
            else getattr(_resource, field_name, None)
        )

        if field_value:
            return field_value

        # Finally check the return_type for mocks

        def resolve_return_type(schema_type: GraphQLType):
            # Special case for list types return a random length list of type.
            if isinstance(schema_type, GraphQLList):
                list_length = mock_objects.get(LIST_LENGTH_KEY, 3)
                return [
                    resolve_return_type(schema_type.of_type) for x in range(list_length)
                ]

            named_type_name = get_named_type(schema_type).name
            if named_type_name in mock_objects:
                return mock_objects.get(named_type_name)

            # If we reached this point this is a custom type that has not
            # explicitly overridden in the mock_objects. Just return a dict
            # with a `__typename` set to the type name to assist in resolving
            # Unions and Interfaces.
            return SuperDict({"__typename": named_type_name})

        return resolve_return_type(return_type)



"""
Debug Middleware
================

Use this middleware to log details about the queries that you are running.

By default this will use logging.debug and will use the cannula logger. You
can override that when you setup the middleware.

Example with Cannula API
------------------------

::

    import cannula
    from cannula.middleware import DebugMiddleware

    api = cannula.API(
        __name__,
        schema=SCHEMA,
        middleware=[
            DebugMiddleware(),
        ],
    )


Example with `graphql-core-next`
--------------------------------

You can optionally use this middleware as a standalone with the `graphql-core-next`::

    from cannula.middleware import DebugMiddleware
    from graphql import graphql

    graphql(
        schema=SCHEMA,
        query=QUERY,
        middleware=[
            DebugMiddleware(),
        ],
    )

"""
import inspect
import logging
import time
import typing

from graphql import GraphQLObjectType, GraphQLResolveInfo


class DebugMiddleware:
    def __init__(self, logger: typing.Optional[logging.Logger] = None):
        self.logger = logger or logging.getLogger("cannula.middleware.debug")

    async def resolve(
        self,
        next_fn: typing.Callable[..., typing.Any],
        parent_object: GraphQLObjectType,
        info: GraphQLResolveInfo,
        **kwargs,
    ):
        parent_name = info.parent_type.name
        field_name = info.field_name
        return_type = info.return_type

        self.logger.debug(
            f"Resolving {parent_name}.{field_name} expecting type {return_type}"
        )

        start_time = time.perf_counter()

        results = next_fn(parent_object, info, **kwargs)

        if inspect.isawaitable(results):
            results = await results

        end_time = time.perf_counter()
        total_time = end_time - start_time
        self.logger.debug(
            f"Field {parent_name}.{field_name} resolved: {results!r} in {total_time:.6f} seconds"
        )

        return results



import dataclasses
import typing


@dataclasses.dataclass
class Argument:
    name: str
    type: typing.Any = None
    value: typing.Any = None
    default: typing.Any = None
    required: bool = False


@dataclasses.dataclass
class Directive:
    name: str
    args: typing.List[Argument]


@dataclasses.dataclass
class Field:
    name: str
    value: str
    func_name: str
    description: typing.Optional[str]
    directives: typing.List[Directive]
    args: typing.List[Argument]
    default: typing.Any = None
    required: bool = False


@dataclasses.dataclass
class FieldType:
    value: typing.Any
    required: bool = False


@dataclasses.dataclass
class ObjectType:
    name: str
    fields: typing.List[Field]
    directives: typing.Dict[str, typing.List[Directive]]
    description: typing.Optional[str] = None


@dataclasses.dataclass
class OperationField:
    name: str
    value: str
    func_name: str
    description: typing.Optional[str]
    directives: typing.List[Directive]
    args: typing.List[Argument]
    default: typing.Any = None
    required: bool = False



from .api import API, Resolver
from .codegen import render_file
from .context import Context, ResolveInfo
from .errors import format_errors
from .schema import build_and_extend_schema, concat_documents, load_schema
from .types import Argument, Directive, Field, FieldType, ObjectType
from .utils import gql

__all__ = [
    "API",
    "Argument",
    "Context",
    "Directive",
    "Field",
    "FieldType",
    "ObjectType",
    "Resolver",
    "ResolveInfo",
    "format_errors",
    "build_and_extend_schema",
    "concat_documents",
    "gql",
    "load_schema",
    "render_file",
]

__VERSION__ = "0.11.0"



import cProfile
import pstats
import io
import inspect
import logging
import typing


class ProfileMiddleware:
    def __init__(
        self,
        level: int = logging.DEBUG,
        logger: typing.Optional[logging.Logger] = None,
    ):
        self.level = level
        self.logger = logger or logging.getLogger(__name__)
        self.profiler = cProfile.Profile()

    async def resolve(self, _next, _resource, _info, **kwargs):
        type_name = _info.parent_type.name  # schema type (Query, Mutation)
        field_name = _info.field_name  # The attribute being resolved

        if type_name not in ["Query", "Mutation"]:
            return await self.run_it(_next, _resource, _info, **kwargs)

        self.logger.log(self.level, f"Profiling {field_name} on {type_name}")

        self.profiler.enable()

        results = await self.run_it(_next, _resource, _info, **kwargs)

        self.profiler.disable()
        s = io.StringIO()
        # TODO(rmyers): Add support for 3.7+
        ps = pstats.Stats(self.profiler, stream=s).sort_stats("cumulative")
        ps.print_stats()
        self.profiler.clear()

        self.logger.log(
            self.level, f"{type_name}: {field_name} profile:\n{s.getvalue()}"
        )

        return results

    async def run_it(self, _next, _resource, _info, **kwargs):
        if inspect.isawaitable(_next):
            results = await _next(_resource, _info, **kwargs)
        else:
            results = _next(_resource, _info, **kwargs)

        if inspect.isawaitable(results):
            return await results
        return results



"""
# Run some stuff
==============

Like a boss
"""
import argparse
import logging
import pathlib
import sys

import cannula

# create the top-level parser for global options
parser = argparse.ArgumentParser(
    prog="cannula",
    description="Cannula cli for general functions",
)
parser.add_argument(
    "--dry_run",
    "--dry-run",
    action="store_true",
    help="do not preform actions",
)
parser.add_argument(
    "--debug",
    "-d",
    action="store_true",
    help="print debug information",
)

# Sub Commands parser
subparsers = parser.add_subparsers(help="sub-command --help", dest="command")  # type: ignore

# create the parser for the "codegen" command
codegen_parser = subparsers.add_parser(
    "codegen",
    help="generate code from the schema",
)
codegen_parser.add_argument(
    "schema",
    help="location of graphql file or directory of files",
)
codegen_parser.add_argument(
    "--dest",
    help="destination to write the file to",
    default="_generated.py",
)


def run_codegen(dry_run: bool, schema: str, dest: str):
    source = pathlib.Path(schema)
    documents = cannula.load_schema(source)
    destination = pathlib.Path(dest)
    cannula.render_file(
        type_defs=documents,
        path=destination,
        dry_run=dry_run,
    )


def main():
    argv = sys.argv[1:] or ["--help"]
    while argv:
        options, argv = parser.parse_known_args(argv)
        if not options.command:
            break
        level = logging.DEBUG if options.debug else logging.INFO
        logging.basicConfig(level=level)
        if options.command == "codegen":
            run_codegen(
                dry_run=options.dry_run,
                schema=options.schema,
                dest=options.dest,
            )



"""
Using the API
=============


"""

import asyncio
import collections
import functools
import logging
import inspect
import pathlib
import typing

from graphql import (
    DocumentNode,
    GraphQLError,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLObjectType,
    execute,
    ExecutionResult,
    GraphQLSchema,
    parse,
    validate_schema,
    validate,
)

from .context import Context
from .schema import (
    build_and_extend_schema,
    load_schema,
    maybe_parse,
)

LOG = logging.getLogger(__name__)

RootType = typing.TypeVar("RootType", covariant=True)


class ParseResults(typing.NamedTuple):
    document_ast: DocumentNode
    errors: typing.List[GraphQLError] = []


class Resolver:
    """
    This class is a helper to organize your project as it grows. It allows you
    to put your resolver modules and schema in different packages. For example::

        app/
            api.py     # `api = cannula.API(args)`
        resolvers/
            books.py   # `books = cannula.Resolver()`
            movies.py  # `movies = cannula.Resolver()`


    You then register resolvers and dataloaders in the same way:

    resolvers/books.py::

        import cannula

        from database.books import get_books

        books = cannula.Resolver()

        @books.query('books')
        async def get_books(source, info, args):
            return await get_books()

    resolvers/moives.py::

        import cannula

        from database.movies import get_movies, fetch_movies_for_book

        movies = cannula.Resolver()

        @movies.query('movies')
        async def get_movies(source, info, args):
            return await get_movies()

        @movies.revolver('Books', 'movies')
        async def list_movies_for_book(book, info) -> list[Movie]:
            return await fetch_movies_for_book(book.id)

    app/api.py::

        import cannula

        from resolvers.books import books
        from resolvers.movies import movies

        api = cannula.API(schema=SCHEMA)
        api.include_resolver(books)
        api.include_resolver(movies)


    """

    registry: typing.Dict[str, dict]

    def __init__(self):
        self.registry = collections.defaultdict(dict)

    def query(self, field_name: typing.Optional[str] = None) -> typing.Any:
        """Query Resolver

        Short cut to add a resolver for a query, by default it will use the
        name of the function as the `field_name` to be resolved::

            resolver = cannula.Resolver()

            @resolver.query
            async def something(parent, info) -> str:
                return "hello world"

            @resolver.query(field_name="something")
            async def some_other_something(parent, info) -> str:
                return "override the function name"

        :param field_name: Field name to resolve, by default the function name will be used.
        """
        return self.resolver("Query", field_name)

    def mutation(self, field_name: typing.Optional[str] = None) -> typing.Any:
        """Mutation Resolver

        Short cut to add a resolver for a mutation, by default it will use the
        name of the function as the `field_name` to be resolved::

            resolver = cannula.Resolver()

            @resolver.mutation
            async def make_it(parent, info, name: str) -> str:
                return "hello world"

            @resolver.mutation(field_name="make_it")
            async def some_other_something(parent, info, name: str) -> str:
                return "override the function name"

        :param field_name: Field name to resolve, by default the function name will be used.
        """
        return self.resolver("Mutation", field_name)

    def resolver(
        self, type_name: str, field_name: typing.Optional[str] = None
    ) -> typing.Any:
        """Field Resolver

        Add a field resolver for a given type, by default it will use the
        name of the function as the `field_name` to be resolved::

            resolver = cannula.Resolver()

            @resolver.resolver("Book")
            async def name(parent, info) -> str:
                return "hello world"

            @resolver.resolver("Book", field_name="something")
            async def some_other_something(parent, info):
                return "override the function name

        :param type_name: Parent object type name that is being resolved.
        :param field_name: Field name to resolve, by default the function name will be used.
        """

        def decorator(function):
            _name = field_name or function.__name__
            self.registry[type_name][_name] = function

        return decorator


class API(typing.Generic[RootType]):
    """
    Your entry point into the fun filled world of graphql. Just dive right in::

        import cannula

        api = cannula.API(schema='''
            extend type Query {
                hello(who: String): String
            }
        ''')

        @api.resolver('Query')
        def hello(who):
            return f'Hello {who}!'

    :param schema: GraphQL Schema for this resolver. This can either be a str or `pathlib.Path` object.
    :param context: Context class to hold shared state, added to GraphQLResolveInfo object.
    :param middleware: List of middleware to enable.
    :param root_value: Mapping of operation names to resolver functions.
    :param kwargs: Any extra kwargs passed directly to graphql.execute function.
    """

    _schema: typing.Union[str, DocumentNode, pathlib.Path]
    _resolvers: typing.List[Resolver]
    _root_value: typing.Optional[RootType]
    _kwargs: typing.Dict[str, typing.Any]

    def __init__(
        self,
        schema: typing.Union[str, DocumentNode, pathlib.Path],
        context: typing.Optional[typing.Any] = None,
        middleware: typing.List[typing.Any] = [],
        root_value: typing.Optional[RootType] = None,
        **kwargs,
    ):
        self._context = context or Context
        self._resolvers = []
        self._schema = schema
        self.middleware = middleware
        self._root_value = root_value
        self._kwargs = kwargs

    def query(self, field_name: typing.Optional[str] = None) -> typing.Any:
        """Query Resolver

        Short cut to add a resolver for a query, by default it will use the
        name of the function as the `field_name` to be resolved::

            api = cannula.API(schema="type Query { something: String }")

            @api.query
            async def something(parent, info):
                return "hello world"

            @api.query(field_name="something")
            async def some_other_something(parent, info):
                return "override the function name"

        :param field_name: Field name to resolve, by default the function name will be used.
        """
        return self.resolver("Query", field_name)

    def mutation(self, field_name: typing.Optional[str] = None) -> typing.Any:
        """Mutation Resolver

        Short cut to add a resolver for a mutation, by default it will use the
        name of the function as the `field_name` to be resolved::

            api = cannula.API(schema="type Mutation { make_it(name: String): String }")

            @api.mutation
            async def make_it(parent, info, name: str):
                return "hello world"

            @api.mutation(field_name="make_it")
            async def some_other_something(parent, info, name: str):
                return "override the function name"

        :param field_name: Field name to resolve, by default the function name will be used.
        """
        return self.resolver("Mutation", field_name)

    def resolver(
        self, type_name: str, field_name: typing.Optional[str] = None
    ) -> typing.Any:
        """Field Resolver

        Add a field resolver for a given type, by default it will use the
        name of the function as the `field_name` to be resolved::

            api = cannula.API(schema="type Book { name: String }")

            @api.resolver("Book")
            async def name(parent, info):
                return "hello world"

            @api.resolver("Book", field_name="something")
            async def some_other_something(parent, info):
                return "override the function name

        :param type_name: Parent object type name that is being resolved.
        :param field_name: Field name to resolve, by default the function name will be used.
        """

        def decorator(function):
            _name = field_name or function.__name__
            field_def = self._validate_field(type_name=type_name, field_name=_name)
            field_def.resolve = function

        return decorator

    def include_resolver(self, resolver: Resolver):
        """Include a set of resolvers

        This is used to break up a larger application into different modules.
        For example you can group all the resolvers for a specific feature set::

            from cannula import Resolver

            from database import Book, filter_books, get_book

            book_resolver = Resolver()

            @book_resolver.query
            async def books(parent, info, **args) -> list[Book]:
                return await filter_books(**args)

            @book_resolver.query
            async def book(parent, info, book_id: str) -> Book:
                return await get_book(book_id)

        Then include the resolver in the main cannula API::

            import cannula
            import pathlib

            from features.books import book_resolver

            api = cannula.API(schema=pathlib.Path('.'))
            api.include_resolver(book_resolver)

        """
        self._merge_registry(resolver.registry)

    def _find_schema(self) -> typing.List[DocumentNode]:
        schemas: typing.List[DocumentNode] = []

        if isinstance(self._schema, pathlib.Path):
            schemas.extend(load_schema(self._schema))
        else:
            schemas.append(maybe_parse(self._schema))

        return schemas

    @property
    def schema(self) -> GraphQLSchema:
        if not hasattr(self, "_full_schema"):
            self._full_schema = self._build_schema()
        return self._full_schema

    def _build_schema(self) -> GraphQLSchema:
        schema = build_and_extend_schema(self._find_schema())

        schema_validation_errors = validate_schema(schema)
        if schema_validation_errors:
            raise Exception(f"Invalid schema: {schema_validation_errors}")

        return schema

    def _validate_field(self, type_name: str, field_name: str) -> GraphQLField:
        object_type = self.schema.get_type(type_name)
        if object_type is None:
            raise Exception(f"Invalid type {type_name}")

        # Need to cast this to object_type to satisfy mypy checks
        object_type = typing.cast(GraphQLObjectType, object_type)
        field_map = typing.cast(GraphQLFieldMap, object_type.fields)
        field_definition = field_map.get(field_name)
        if not field_definition:
            raise Exception(f"Invalid field {type_name}.{field_name}")

        return field_definition

    def context(self):
        def decorator(klass):
            self._context = klass

        return decorator

    def get_context(self, request) -> typing.Any:
        return self._context.init(request)

    def _merge_registry(self, registry: dict):
        for type_name, value in registry.items():
            for field_name, revolve_fn in value.items():
                field_def = self._validate_field(type_name, field_name)
                field_def.resolve = revolve_fn

    @functools.lru_cache(maxsize=128)
    def _validate(self, document: DocumentNode) -> typing.List[GraphQLError]:
        """Validate the document against the schema and store results in lru_cache."""
        return validate(self.schema, document)

    @functools.lru_cache(maxsize=128)
    def _parse_document(self, document: str) -> ParseResults:
        """Parse and store the document in lru_cache."""
        try:
            document_ast = parse(document)
            return ParseResults(document_ast, [])
        except GraphQLError as err:
            return ParseResults(DocumentNode(), [err])

    async def call(
        self,
        document: typing.Union[DocumentNode, str],
        request: typing.Any = None,
        variables: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> ExecutionResult:
        """Preform a query against the schema.

        This is meant to be called in an asyncio.loop, if you are using a
        web framework that is synchronous use the `call_sync` method.
        """
        if isinstance(document, str):
            document, errors = self._parse_document(document)
            if errors:
                return ExecutionResult(data=None, errors=errors)

        if validation_errors := self._validate(document):
            return ExecutionResult(data=None, errors=validation_errors)

        context = self.get_context(request)
        result = execute(
            schema=self.schema,
            document=document,
            context_value=context,
            variable_values=variables,
            middleware=self.middleware,
            root_value=self._root_value,
            **self._kwargs,
        )
        if inspect.isawaitable(result):
            return await result
        return typing.cast(ExecutionResult, result)

    def call_sync(
        self,
        document: DocumentNode,
        request: typing.Optional[typing.Any] = None,
        variables: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> ExecutionResult:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.call(document, request, variables))
